#include <engine/psx.h>
#include <engine/math.h>
#include <limits.h>

/*
 * psqrt:
 *	Uses binary search to take the square root of
 *	a Q10.6 number. For Q12.4 numbers, use csqrt().
 *	Recursive square root approaches one.
 */
unsigned int
psqrt(num)
unsigned int num;
{
	int		i;
	unsigned short	max_guess, guess, delta;
	unsigned int	estimate = 0;

	if(num == 1<<FIXED_PRECISION)
	{
		return 1<<FIXED_PRECISION;
	}

	/*
	 * prevent overflow, since 65536^2 == 2^32, which is greater
	 * than UINT_MAX.
	 */
	//if(num < )
	max_guess = num > USHRT_MAX ? USHRT_MAX : num;

	guess = delta = max_guess>>1;

#define MAX_ITERATIONS 32

	for(i = 0; i < MAX_ITERATIONS; i++)
	{
		if(!delta)
		{
			return guess;
		}

		estimate = (guess*guess)>>FIXED_PRECISION;

		if(estimate > num)
		{
			guess -= delta;
		}
		else if(estimate < num)
		{
			guess += delta;
		}
		else
		{
			return guess;
		}

		delta >>= 1;
	}

	return guess;
}

/*
 * psqrt2:
 *	Uses binary search to take the square root of
 *	any number.
 */
unsigned int
psqrt2(num, precision)
unsigned int num;
unsigned int precision;
{
	int		i;
	unsigned short	max_guess, guess, delta;
	unsigned int	estimate = 0;

	if(num == 1<<precision)
	{
		return 1<<precision;
	}
	/*
	 * prevent overflow, since 65536^2 == 2^32, which is greater
	 * than UINT_MAX.
	 */
	max_guess = num > USHRT_MAX ? USHRT_MAX : num;

	guess = delta = max_guess>>1;

#define MAX_ITERATIONS 32

	for(i = 0; i < MAX_ITERATIONS; i++)
	{
		if(!delta)
		{
			return guess;
		}

		estimate = (guess*guess)>>precision;

		if(estimate > num)
		{
			guess -= delta;
		}
		else if(estimate < num)
		{
			guess += delta;
		}
		else
		{
			return guess;
		}

		delta >>= 1;
	}

	return guess;
}

/*
 * sign:
 *	Returns the sign of a number.
 */
sign(num)
int num;
{
	if(num < 0)
	{
		return -1;
	}

	return 1;
}

/*
 * lerp:
 *	Linearly interpolate between two points.
 *	a and b are normal integers, but t is a Q10.6
 *	number. The number returned is in Q10.6 format.
 */
lerp(a, b, t)
int a, b, t;
{
	return (((1<<6)-t)*a) + (b*t);
}

/*
 * lerpi:
 *	Linearly interpolate between two points.
 *	a and b are normal integers, but Q dictates
 *	the number of fractional bits t has.
 */
lerpi(a, b, t, Q)
int a, b, t, Q;
{
	return (((1<<Q)-t)*a) + (b*t);
}

/*
 * plerp:
 *	Linearly interpolate between two points.
 *	All arguments and return value are in Q10.6.
 */
plerp(a, b, t)
int a, b, t;
{
	return ((((1<<6)-t)*a)>>6) + ((b*t)>>6);
}

/*
 * ppow:
 *	Calculates a base taken to the power of an exponent.
 */
ppow(base, exponent)
int base, exponent;
{
	int i, result = base;

	if(!exponent)
	{
		return 1;
	}

	for(i = 2; i <= exponent; i++)
	{
		result *= base;
	}
	return result;
}

/*
 * pclamp:
 *	Clamps a value between a min and max.
 */
pclamp(num, min, max)
{
	if(num < min)
	{
		return min;
	}
	else if(num > max)
	{
		return max;
	}
	return num;
}

/*
 * pdot:
 *	Calculates the dot product of two vectors.
 */
pdot(p_a, p_b)
VECTOR *p_a, *p_b;
{
	return (p_a->vx*p_b->vx + p_a->vy*p_b->vy + p_a->vz*p_b->vz)>>FIXED_PRECISION;
}

/*
 * pcross:
 *	Calculates the cross product of two vectors.
 */
pcross(p_a, p_b, Q, p_result)
VECTOR *p_a, *p_b;
int Q;
VECTOR *p_result;
{
	p_result->vx = ( ((p_a->vy*p_b->vz)>>Q) - ((p_a->vz*p_b->vy)>>Q));
	p_result->vy = ( ((p_a->vz*p_b->vx)>>Q) - ((p_a->vx*p_b->vz)>>Q));
	p_result->vz = ( ((p_a->vx*p_b->vy)>>Q) - ((p_a->vy*p_b->vx)>>Q));
}

/*
 * quat_conj:
 *	Derives the conjugate of a quaternion. Stores the result into
 *	p_quat_b.
 */
void
quat_conj(p_quat_a, p_quat_b)
struct Quaternion *p_quat_a, *p_quat_b;
{
	p_quat_b->s = p_quat_a->s;
	p_quat_b->i = -p_quat_a->i;
	p_quat_b->j = -p_quat_a->j;
	p_quat_b->k = -p_quat_a->k;
}

/*
 * quat_mult:
 *	Multiplies two quaternions. Stores the result into
 *	p_quat_c.
 *
 *	[s0s1 - v0*v1, s0v1 + s1v0 + v0xv1]
 */
void
quat_mult(p_quat_a, p_quat_b, p_quat_c)
struct Quaternion *p_quat_a, *p_quat_b, *p_quat_c;
{
	VECTOR vec;

	p_quat_c->s = ((p_quat_a->s*p_quat_b->s)>>FIXED_PRECISION) - pdot((VECTOR *)(&p_quat_a->i), (VECTOR *)(&p_quat_b->i));

	p_quat_c->i = (p_quat_a->s*p_quat_b->i)>>FIXED_PRECISION;
	p_quat_c->j = (p_quat_a->s*p_quat_b->j)>>FIXED_PRECISION;
	p_quat_c->k = (p_quat_a->s*p_quat_b->k)>>FIXED_PRECISION;

	p_quat_c->i += (p_quat_b->s*p_quat_a->i)>>FIXED_PRECISION;
	p_quat_c->j += (p_quat_b->s*p_quat_a->j)>>FIXED_PRECISION;
	p_quat_c->k += (p_quat_b->s*p_quat_a->k)>>FIXED_PRECISION;

	pcross((VECTOR *)(&p_quat_a->i), (VECTOR *)(&p_quat_b->i), FIXED_PRECISION, &vec);

	p_quat_c->i += vec.vx;
	p_quat_c->j += vec.vy;
	p_quat_c->k += vec.vz;
}

/*
 * quat_mag:
 *	Returns the magnitude of a quaternion (which is a scalar!)
 */
unsigned int
quat_mag(p_quat_a)
struct Quaternion *p_quat_a;
{
	return psqrt(((p_quat_a->s*p_quat_a->s)>>FIXED_PRECISION) +
			((p_quat_a->i*p_quat_a->i)>>FIXED_PRECISION) +
			((p_quat_a->j*p_quat_a->j)>>FIXED_PRECISION) +
			((p_quat_a->k*p_quat_a->k)>>FIXED_PRECISION));
}

/*
 * quat_norm:
 *	Normalizes a quaternion. FIXME: can be more accurate by rounding!
 */
void
quat_norm(p_quat_a)
struct Quaternion *p_quat_a;
{
	int norm;

	norm = quat_mag(p_quat_a);
	norm = !norm ? 1 : norm;

	p_quat_a->s = (p_quat_a->s<<FIXED_PRECISION)/norm;
	p_quat_a->i = (p_quat_a->i<<FIXED_PRECISION)/norm;
	p_quat_a->j = (p_quat_a->j<<FIXED_PRECISION)/norm;
	p_quat_a->k = (p_quat_a->k<<FIXED_PRECISION)/norm;
}

/*
 * quat_inv:
 *	Determines the inverse of a quaternion. Stores the result
 *	into p_quat_b.
 */
void
quat_inv(p_quat_a, p_quat_b)
struct Quaternion *p_quat_a, *p_quat_b;
{
	struct Quaternion conj;
	int norm, norm2, normrec;

	quat_conj(p_quat_a, &conj);
	norm = quat_mag(p_quat_a);
	//printf("\n----quat_inv----\nnorm: %d\n", norm);
	//while(1);
	norm2 = ((norm*norm)>>FIXED_PRECISION);
	norm2 = !norm2 ? 1 : norm2;

	normrec = (FIXED_PRECISION*FIXED_PRECISION)/norm2; /* reciprocal of norm squared */

	/* conjugate over the square of the norm ( q* / |q|^2 ) */

	//printf("\n------quat_inv:\n");
	//printf("quat: %d, %d, %d, %d\n", p_quat_a->s, p_quat_a->i, p_quat_a->j, p_quat_a->k);
	//printf("norm: %d\n", norm);
	//printf("norm2: %d\n", norm2);
	//printf("conj: %d, %d, %d, %d\n", conj.s, conj.i, conj.j, conj.k);
	//printf("\n");

	p_quat_b->s = (conj.s<<FIXED_PRECISION)/norm2;
	p_quat_b->i = (conj.i<<FIXED_PRECISION)/norm2;
	p_quat_b->j = (conj.j<<FIXED_PRECISION)/norm2;
	p_quat_b->k = (conj.k<<FIXED_PRECISION)/norm2;

	///* conjugate over the square of the norm ( q* / |q|^2 ) */
	//p_quat_b->s = (conj.s*normrec)>>FIXED_PRECISION;
	//p_quat_b->i = (conj.i*normrec)>>FIXED_PRECISION;
	//p_quat_b->j = (conj.j*normrec)>>FIXED_PRECISION;
	//p_quat_b->k = (conj.k*normrec)>>FIXED_PRECISION;
}

/*
 * quat_rot:
 *	Derives a rotation quaternion. Stores the result into
 *	p_quat_b.
 */
void
quat_rot(p_vec, theta, p_quat_b)
VECTOR *p_vec; /* v-hat */
int theta;
struct Quaternion *p_quat_b;
{
	int sinhalf = rsin(theta)>>(FIXED_PRECISION+1);

	p_quat_b->s = rcos(theta)>>(FIXED_PRECISION+1);
	p_quat_b->i = (sinhalf*p_vec->vx)>>FIXED_PRECISION;
	p_quat_b->j = (sinhalf*p_vec->vy)>>FIXED_PRECISION;
	p_quat_b->k = (sinhalf*p_vec->vz)>>FIXED_PRECISION;
}

/*
 * vec_mag:
 *	Returns the magnitude of a vector.
 */
unsigned int
vec_mag(p_vec)
VECTOR *p_vec;
{
	return psqrt(((p_vec->vx*p_vec->vx)>>FIXED_PRECISION) +
		((p_vec->vy*p_vec->vy)>>FIXED_PRECISION) +
		((p_vec->vz*p_vec->vz)>>FIXED_PRECISION));
}

/*
 * vec_norm:
 *	Normalizes a vector.
 */
void
vec_norm(p_vec)
VECTOR *p_vec;
{
	int r = vec_mag(p_vec);

	r = !r ? 1 : r;

	p_vec->vx = ((p_vec->vx<<6)+32)/r;
	p_vec->vy = ((p_vec->vy<<6)+32)/r;
	p_vec->vz = ((p_vec->vz<<6)+32)/r;
}

/*
 * vec_ortho:
 *	Finds an arbitrary vector that is orthogonal to another vector.
 */
void
vec_ortho(p_vec_a, p_vec_b)
VECTOR *p_vec_a, *p_vec_b;
{
	int i;
	int comp[3] = {0, 0, 0};
	int ncomp = 0;

	int *p_vec_a_int = p_vec_a;
	int *p_vec_b_int = p_vec_b;

	{
		int idx = 0;
		for(i = 0; i < 3; i++)
		{
			if(p_vec_a_int[i] != 0)
			{
				comp[idx] = i, idx++;
				ncomp++;
			}
		}
	}

	if(!ncomp)
	{
		printf("vec_ortho() error: input is a zero vector: (%d, %d, %d)\n", p_vec_a->vx, p_vec_a->vy, p_vec_a->vz);
		return;
	}

	{
		*p_vec_b = *p_vec_a;

		switch(ncomp)
		{
			case 1:
				for(i = 0; i < 3; i++)
				{
					if(p_vec_b_int[i])
					{
						p_vec_b_int[(i+1)%3] = p_vec_b_int[i];
						p_vec_b_int[i] = 0;
						break;
					}
				}
				break;
			case 2:
				p_vec_b_int[comp[0]] = p_vec_a_int[comp[1]];
				p_vec_b_int[comp[1]] = -p_vec_a_int[comp[0]];
				break;
			case 3:
				setVector(p_vec_b, 0, p_vec_a->vz, -p_vec_a->vy);
				break;
			default:
				printf("vec_ortho() error: Number of components is either 0 or exceeds 3 (ncomp == %d)\n", ncomp);
				break;
		}
	}
}



///*
// * vec_add:
// *	Adds two vectors.
// */
//void
//vec_add(p_vec_a, p_vec_b, p_result)
//VECTOR *p_vec_a, *p_vec_b, *p_result;
//{
//	p_result->vx = p_vec_a->vx + p_vec_b->vx;
//	p_result->vy = p_vec_a->vy + p_vec_b->vy;
//	p_result->vz = p_vec_a->vz + p_vec_b->vz;
//}

///*
// * vec_sub:
// *	Subtracts two vectors.
// */
//void
//vec_sub(p_vec_a, p_vec_b, p_result)
//VECTOR *p_vec_a, *p_vec_b, *p_result;
//{
//	p_result->vx = p_vec_a->vx - p_vec_b->vx;
//	p_result->vy = p_vec_a->vy - p_vec_b->vy;
//	p_result->vz = p_vec_a->vz - p_vec_b->vz;
//}

///*
// * vec_mult:
// *	Multiplies two vectors. Assumes FIXED_PRECISION.
// */
//void
//vec_mult(p_vec_a, p_vec_b, p_result)
//VECTOR *p_vec_a, *p_vec_b, *p_result;
//{
//	p_result->vx = (p_vec_a->vx * p_vec_b->vx)>>FIXED_PRECISION;
//	p_result->vy = (p_vec_a->vy * p_vec_b->vy)>>FIXED_PRECISION;
//	p_result->vz = (p_vec_a->vz * p_vec_b->vz)>>FIXED_PRECISION;
//}

///*
// * vec_div:
// *	Divides two vectors. Assumes FIXED_PRECISION.
// */
//void
//vec_div(p_vec_a, p_vec_b, p_result)
//VECTOR *p_vec_a, *p_vec_b, *p_result;
//{
//	p_result->vx = (p_vec_a->vx / p_vec_b->vx)<<FIXED_PRECISION;
//	p_result->vy = (p_vec_a->vy / p_vec_b->vy)<<FIXED_PRECISION;
//	p_result->vz = (p_vec_a->vz / p_vec_b->vz)<<FIXED_PRECISION;
//}

/*
 * closest_point_on_line:
 *	Finds the closest point to p_vec that is contained within the line
 *	defined by points A and B.
 */
void
closest_point_on_line(p_a, p_b, p_vec, p_result)
VECTOR *p_a, *p_b, *p_vec, *p_result;
{
	VECTOR ab, tmp;
	int t;

	vec_sub(p_b, p_a, &ab);
	vec_sub(p_vec, p_a, &tmp);
	t = pclamp((pdot(&tmp, &ab)/pdot(&ab, &ab))<<FIXED_PRECISION, 0, 1<<FIXED_PRECISION);

	ab.vx = (ab.vx*t)>>FIXED_PRECISION;
	ab.vy = (ab.vy*t)>>FIXED_PRECISION;
	ab.vz = (ab.vz*t)>>FIXED_PRECISION;
	vec_add(p_a, &ab, p_result);
}

